/**
 ** Simple entropy harvester based upon the havege RNG
 **
 ** Copyright 2009-2012 Gary Wuertz gary@issiweb.com
 ** Copyright 2011-2012 BenEleventh Consulting manolson@beneleventh.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef HAVEGETEST_H
#define HAVEGETEST_H
/**
 * The haveged test suite is built from the 8 tests specified in AIS-31
 * organized into test procedure A and test procedure B and structured
 * as state machines capable of processing segmented input streams.
 */
#include <stdint.h>
/**
 * Serious bit twiddling requires serious integer type precision
 */
#define  U_INT32  uint_least32_t
#define  U_INT48  long long
#define  U_INT8   uint8_t
/**
 * All individual tests and the test procedures use the following
 * simple state maching to manage input.
 */
typedef enum {
  TEST_INIT,               /* initialize the test           */
  TEST_INPUT,              /* test input needed             */
  TEST_EVAL,               /* evaluating results (not used) */
  TEST_DONE,               /* test complete                 */
  TEST_RETRY               /* retry the test                */
  } TEST_STATE;
/**
 * AIS-31 procedure A uses the FIPS140-1 as test1 thru test4. A disjointedness test is
 * used as test0 and a autocorrelation test is used as test5. test0 is executed only
 * once, the other tests are repeated in sequence 257 times.
 */
#define  AIS_A_REPS        257   /* reps for test1 thru test 5       */
/**
 * Constants for the fips tests. Note AIS-31 v1 uses the unamended FIPS test limits
 */
#define  FIPS_LENGTH       (20000>>3)
#ifndef  USE_AMENDED_FIPS
#define  FIPS_MAX_RUN      34
#define  FIPS_ONES_LOW     9654
#define  FIPS_ONES_HIGH    10346
#define  FIPS_POKER_LOW    1562822  // 1.03
#define  FIPS_POKER_HIGH   1580438  // 57.4
#define  FIPS_RUNS_LOW     2267,1079,502,223,90,90
#define  FIPS_RUNS_HIGH    2733,1421,748,402,223,223
#else
#define  FIPS_MAX_RUN      25
#define  FIPS_ONES_LOW     9725
#define  FIPS_ONES_HIGH    10275
#define  FIPS_POKER_LOW    1576928  // 2.16
#define  FIPS_POKER_HIGH   1576928  // 46.17
#define  FIPS_RUNS_LOW     2315,1114,525,240,103,103
#define  FIPS_RUNS_HIGH    2685,1386,723,384,209,209
#endif
/**
 * test 0 consumes 64k * 48 bits
 */
#define  TEST0_LENGTH      65536
#define  TEST0_SIZE        (TEST0_LENGTH*sizeof(U_INT48))
#define  TEST0_USED        (TEST0_LENGTH * 6)
#define  TEST5_LENGTH      5000
/**
 * Fixed size input for procedure A
 */
#define  AIS_A_SIZE  (TEST0_USED+(2500*257))
/**
 * AIS-31 procedure A results
 */
typedef struct {
   U_INT testResult;             /* id 8 bits, pass/fail 8bits */
   U_INT finalValue;             /* end result                 */
} resultA;
/**
 * AIS-31 procedure A context. Options are defined in haveged.h
 */
typedef struct {
   char     *data;               /* input for test             */         
   U_INT    range;               /* number of words of input   */
   U_INT    procState;           /* procedure state            */
   U_INT    procRetry;           /* retry indication           */
   U_INT    testId;              /* test selector 0-5          */
   U_INT    testRun;             /* test index 1 - 1285        */
   U_INT    testState;           /* FSM state of current test  */
   U_INT    bridge;              /* index for data bridge      */
   U_INT    bytesUsed;           /* number of bytes used       */
   U_INT8   *chain5;             /* chain to test 5            */
   U_INT    options;             /* duty cycle for test5       */
   char     aux[TEST0_SIZE];     /* extra work space           */
   resultA  results[1286];       /* test results               */
   } procA;
/**
 * AIS-31 procedure B is a set of multinomial distribution tests
 * and an entropy estimate (Coron' test). The distribution tests,
 * test6 and test7, require at least AIS_LENGTH sequences of 1, 2
 * 4, and 8 bits.
 */
/*
 * Bit range of AIS-31 procedure B distribution tests
 */
#define  AIS_LENGTH     100000
/**
 * AIS-31 test8 constants (Coron's test)
 */
#define  Q  2560
#define  K  256000
#define LN2 0.69314718055994530941
/**
 * AIS-31 procedure B results
 */
typedef struct {
   U_INT testResult;             /* id 8 bits, pass/fail 8bits */
   double finalValue;            /* final value                */
   } resultB;
/**
 * AIS-31 procedure B context
 */
typedef struct {
   U_INT    *noise;              /* input for test             */         
   U_INT    range;               /* number of words of input   */
   U_INT    procState;           /* procedure state            */
   U_INT    procRetry;           /* retry indication           */
   U_INT    testId;              /* test selector 6-8          */ 
   U_INT    testNbr;             /* current test number        */
   U_INT    testState;           /* FSM state of current test  */
   U_INT    seq;                 /* aisSeq() sequence needed   */
   U_INT    bitsUsed;            /* bits used by procedure     */
   U_INT    bridge;              /* data bridge test6,7        */
   U_INT    counter[8];          /* sequence lengths           */
   U_INT    einsen[8];           /* sequence counts (ones)     */
   U_INT    full;                /* sequence flags             */
   U_INT    lastpos[256];        /* counters for test 8        */
   U_INT    options;             /* RFU                        */
   resultB  results[9];          /* test results               */
   } procB;
/**
 * A test procedure run consists of an indicator and options
 */
typedef struct {
   U_INT    action;              /* action code A_RUN, B_RUN  */
   U_INT    options;             /* WARN and other options    */
   } procInst;
/**
 * A test procedure is associated with a collection buffer. Some
 * resources are shared by all collection buffers
 */
typedef struct {
   U_INT    options;             /* verbosity, etc.            */
   U_INT    testsUsed;           /* tests used                 */
   procInst totTests[2];         /* tot tests to run           */
   procInst runTests[2];         /* production tests to run    */
   U_INT    procReps;            /* Number of  A repetitions   */
   U_INT    fips_low[6];         /* low runs thresholds        */
   U_INT    fips_high[6];        /* high runs thresholds       */
   double   *G;                  /* test8 lookup table         */
} procShared;
/**
 * Online testing context
 */
typedef struct {
   U_INT       result;           /* nz if failed               */
   U_INT       totIdx;           /* tot test idx               */
   U_INT       runIdx;           /* run test idx               */
   U_INT       szTotal;          /* total bytes processed      */
   procA       *pA;              /* procedure A instance       */
   procB       *pB;              /* procedure B instance       */
} onlineTests;

#endif
